from __future__ import print_function
__author__ = 'marco'

import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import euclidean

import traceback
import hierarchicalDBSCAN.Clash.toolbox as toolbox
import math
import hierarchicalDBSCAN.Clash.statistics as statistics


def get_concave_hull(ps, initial_hull_edges=None, termination_criterion='delauney', initial_termination_threshold=None,
                     percentile=0.75, perc_multiplier=3, adaptive_termination=True, stop_at_disjoin=False,
                     plot_steps=False, plot_steps_after=0, verbose=False, counter=None):

    if len(ps) < 3:
        if verbose:
            print('subset too small to process.')
        return [], counter
    elif verbose:
        print()

    current_hull_edges = initial_hull_edges if initial_hull_edges is not None else toolbox.get_convex_hull(ps)

    if initial_termination_threshold is not None:
        edge_length_threshold = initial_termination_threshold
    else:
        edge_length_threshold = statistics.get_nn_edge_length_noise_threshold(ps, percentile * 100, perc_multiplier, verbose) \
                                if termination_criterion == 'nearest_neighbor' else \
                                statistics.get_delaunay_edge_length_noise_threshold(ps, percentile * 100, perc_multiplier, verbose)

    longest_edge, longest_edge_length = toolbox.get_longest_edge(ps, current_hull_edges, 0)
    counter = 0 if counter is None else counter

    if plot_steps:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    while longest_edge_length > edge_length_threshold:

        if verbose:
            print('iteration ' + str(counter), end='')

        old_hull_edges = current_hull_edges
        success, current_hull_edges = get_concave_hull__step(ps,
                                                             current_hull_edges,
                                                             plot_step=(plot_steps and counter >= plot_steps_after),
                                                             verbose=verbose)
        disjointed_hulls_edges, disjointed_pss, was_disjoined = disjoin_if_necessary(current_hull_edges, ps)
        if was_disjoined and stop_at_disjoin:
            return [old_hull_edges], [ps]

        if not success:
            break
        elif len(disjointed_hulls_edges) == 0:
            return [], counter
        elif len(disjointed_hulls_edges) > 1:
            final_hulls_edge_lists = []
            if verbose:
                print('splitting into ' + str(len(disjointed_hulls_edges)) + ' subsets...')
            for disj_hull, disj_ps in zip(disjointed_hulls_edges, disjointed_pss):
                if not adaptive_termination:
                    map = np.asarray(toolbox.map_point_set(disj_ps, ps))
                    disj_hull = [(np.argwhere(map==edge[0])[0][0], np.argwhere(map==edge[1])[0][0]) for edge in disj_hull]
                disj_hulls, counter = get_concave_hull(ps=disj_ps,
                                              initial_hull_edges=None if adaptive_termination else disj_hull,
                                              initial_termination_threshold=None if adaptive_termination else edge_length_threshold,
                                              percentile=percentile, perc_multiplier=perc_multiplier,
                                              adaptive_termination=adaptive_termination,
                                              plot_steps=plot_steps, plot_steps_after=plot_steps_after, verbose=verbose, counter=counter)
                for disj_hull in disj_hulls:
                    disj_hull_mapped = toolbox.map_edge_list(disj_hull, toolbox.map_point_set(disj_ps, ps))
                    final_hulls_edge_lists.append(disj_hull_mapped)
            return final_hulls_edge_lists, counter

        else:
            current_hull_edges = disjointed_hulls_edges[0]
            if len(current_hull_edges) < 3:
                return [], counter
            _, longest_edge_length = toolbox.get_longest_edge(ps, current_hull_edges, 0)
            counter += 1

    if verbose:
        print()
        print("done after " + str(counter) + " iterations.")

    return [current_hull_edges], counter






def get_concave_hull__step(ps, current_hull_edges, plot_step=False, verbose=False):

    skip_counter = 0
    while skip_counter < len(current_hull_edges):
        longest_edge, longest_edge_length = toolbox.get_longest_edge(ps, current_hull_edges, rank=skip_counter)
        if verbose:
            print(' | ' + str(skip_counter) + '. longest edge: ' + str(longest_edge) + ' (' + str(longest_edge_length) + ')', end='')

        if plot_step:
            current_hull_edges_with_last_edge = current_hull_edges[:]
            current_hull_edges_with_last_edge.append((current_hull_edges[-1][1], current_hull_edges[0][0]))
            current_hull_edges_with_last_edge = np.asarray(current_hull_edges_with_last_edge)
            plt.plot(ps[current_hull_edges_with_last_edge[:, 0]][:, 0], ps[current_hull_edges_with_last_edge[:, 0]][:, 1], 'b-')


        try:
            split_index, split_point, left_pts, right_pts = split_point_set(ps, longest_edge, plot_step)
        except ValueError as e:
            skip_counter += 1
            print(" - value error - ")
            continue

        if verbose:
            print(" >> splitting at " + str(split_index), end="")

        if plot_step:
            not_left_right_pts = [idx for idx in range(0, len(ps)) if (idx not in left_pts[0]) and (idx not in right_pts[0])]
            plt.plot(ps[left_pts][:, 0], ps[left_pts][:, 1], 'go', alpha=0.125)
            plt.plot(ps[right_pts][:, 0], ps[right_pts][:, 1], 'yo', alpha=0.125)
            plt.plot(ps[not_left_right_pts][:, 0], ps[not_left_right_pts][:, 1], 'ko', alpha=0.125)
            plt.plot(ps[longest_edge[0]][0], ps[longest_edge[0]][1], 'rx', markersize=5)
            plt.plot(ps[longest_edge[1]][0], ps[longest_edge[1]][1], 'gx', markersize=5)
            plt.plot(ps[split_index][0], ps[split_index][1], 'gx', markersize=5)
            plt.axis('equal')
            plt.show()

        left_hull_edges = toolbox.get_convex_hull(ps[left_pts])
        right_hull_edges = toolbox.get_convex_hull(ps[right_pts])

        if verbose:
            print(' >> shrinking hull...', end='')
        if left_hull_edges is None:
            if split_index in toolbox.edge_list_to_hull(current_hull_edges):
                current_hull_edges = toolbox.subchain(current_hull_edges, longest_edge[1], split_index)
                current_hull_edges.append((split_index, longest_edge[1]))
            else:
                last_edge = toolbox.subchain(current_hull_edges, longest_edge[1], longest_edge[0])[-1]
                current_hull_edges = toolbox.subchain(current_hull_edges, longest_edge[1], longest_edge[0])[:-1]
                current_hull_edges.append((last_edge[0], split_index))
                current_hull_edges.append((split_index, longest_edge[1]))
        elif right_hull_edges is None:
            if split_index in toolbox.edge_list_to_hull(current_hull_edges):
                current_hull_edges = toolbox.subchain(current_hull_edges, split_index, longest_edge[0])
                current_hull_edges.append((longest_edge[0], split_index))
            else:
                first_edge = toolbox.subchain(current_hull_edges, longest_edge[1], longest_edge[0])[0]
                current_hull_edges = toolbox.subchain(current_hull_edges, longest_edge[1], longest_edge[0])[1:]
                current_hull_edges.append((longest_edge[0], split_index))
                current_hull_edges.append((split_index, first_edge[1]))
        else:
            left_pts = left_pts[0]
            right_pts = right_pts[0]
            left_hull_edges = [(left_pts[edge[0]], left_pts[edge[1]]) for edge in left_hull_edges]
            right_hull_edges = [(right_pts[edge[0]], right_pts[edge[1]]) for edge in right_hull_edges]
            current_hull_edges = shrink_hull(current_hull_edges, left_hull_edges, right_hull_edges, split_index, longest_edge)

        if verbose:
            print(' >> done shrinking hull.')

        return True, current_hull_edges

    return False, current_hull_edges






# longest_edge ... like [[p1, p2], [p2, p3], ...] with p1, ... being int (indices in the original point set)
# returns      ... split_index, split_point, left_pts_indices, right_pts_indices
#   split_index       ... index of the point where split was done
#   split_point       ... coordinates of the point where split was done
#   left_pts_indices  ... indices in the original point set of the points in the left point set
#   right_pts_indices ... analogously
def split_point_set(ps, longest_edge, plot_steps=False):

    # ps[longest_edge, :] like [[x1, y1], [x2, y2]]
    mid_point = np.sum(ps[longest_edge, :], axis=0)/2  # mid_point like [ (x1+x2)/2 , (y1+y2)/2 ]

    slope = ps[longest_edge[1]] - ps[longest_edge[0]]  # slope like [x2-x1, y2-y1]
    perp = -np.asarray([-slope[1], slope[0]])
    perp /= np.linalg.norm(perp)
    new_point = mid_point - perp

    if plot_steps:
        plt.plot([ps[longest_edge[0]][0], ps[longest_edge[1]][0]], [ps[longest_edge[0]][1], ps[longest_edge[1]][1]], 'r:', linewidth=2.0)
        plt.plot(mid_point[0], mid_point[1], 'cD')
        plt.plot(new_point[0], new_point[1], 'cD')
        plt.plot([mid_point[0], new_point[0]], [mid_point[1], new_point[1]], 'r:', markevery=1, marker='^', linewidth=2.0)

    perp_norm = -np.asarray([-perp[1], perp[0]])
    xs = ps - mid_point  # shift dataset so mid_point is the origin

    # determine all points "above" the longest edge (in direction of perp)
    dps_perp = np.asarray([np.dot(x, perp) for x in xs])
    relevant_points = dps_perp <= -0.0000000001

    # determine smallest angle by finding point with smallest dot product
    #dps_perp_norm = np.asarray([np.dot(x, perp_norm) for x in xs])
    #idx = np.argmin([np.abs(d) if d >= 0 else np.inf for d in dps_perp_norm])  # get point closest to normal vector on one side
    #idx2 = np.argmin([np.abs(d) if d <= 0 else np.inf for d in dps_perp_norm])  # get point closest to normal vector on other side

    # determine top 5% of all points with smallest angle.
    # Closest to longest edge will be split point.
    dps_perp_norm = np.asarray([np.dot(x, perp_norm) for x in xs])
    dps_perp_norm_threshold_best5 = np.percentile(np.absolute(dps_perp_norm), 5)
    dps_perp_norm_threshold_all = euclidean(ps[longest_edge[0]], ps[longest_edge[1]]) / 2 - 0.0000000001
    dps_perp_norm_threshold = min(dps_perp_norm_threshold_best5, dps_perp_norm_threshold_all)

    #best_points = np.logical_and(-dps_perp_norm_threshold <= dps_perp_norm, dps_perp_norm <= dps_perp_norm_threshold)
    #best_points = np.logical_and(best_points, relevant_points)
    #best_idx = np.nonzero(best_points)[0]
    # if best 5 % is empty, try all points 'above' line within its range
    #if len(best_idx) == 0:
    #best_points = np.abs(dps_perp_norm) <= dps_perp_norm_threshold_all
    best_points = np.logical_and(-dps_perp_norm_threshold_all <= dps_perp_norm, dps_perp_norm <= dps_perp_norm_threshold_all)
    best_points = np.logical_and(best_points, relevant_points)
    best_idx = np.nonzero(best_points)[0]

    best_lengths = [euclidean(ps[idx], mid_point) for idx in best_idx]
    split_idx = best_idx[np.argmin(best_lengths)]

    if plot_steps:
        plt.plot(ps[best_idx, 0], ps[best_idx, 1], 'gv', alpha=0.5, markersize=5)
        plt.plot(ps[split_idx, 0], ps[split_idx, 1], 'rv', alpha=0.75, markersize=5)

    # determine distance of split_point -> perp, and shift dataset accordingly
    #dist_split_point_to_perp = np.dot(perp_norm, ps[split_idx])

    # determine all points left / right of perpendicular line from split_point -> longest_edge
    #split_vec = ps[split_idx] - mid_point
    xs = xs - xs[split_idx]
    #split_vec_norm = np.asarray([-split_vec[1], split_vec[0]])
    dps_split_norm = np.asarray([np.dot(x, perp_norm) for x in xs])

    left_points = dps_split_norm >= 0 - 0.0000000001  # like [False, False, True, ...]
    right_points = dps_split_norm <= 0 + 0.0000000001
    left_relevant_points = np.logical_and(relevant_points, left_points)
    right_relevant_points = np.logical_and(relevant_points, right_points)

    left_relevant_points[split_idx] = right_relevant_points[split_idx] = True
    # bugfix:
    left_relevant_points[longest_edge[0]] = True
    right_relevant_points[longest_edge[1]] = True

    return split_idx, ps[split_idx], np.nonzero(left_relevant_points), np.nonzero(right_relevant_points)



#####################
# Hull Manipulation #
#####################


# parameters:
#  - edge_list ... list of edges [(idx1, idx2), ...] with idx1, ... being indices in the point set ps
#  - ps        ... a set of points like [(coord_x, coord_y), ...]
# returns:
#  - list of edge lists
#  - list of ps sets, each set containing all points within the borders of the corresponding edge list
#  - was_disjoined ... True, if there was a duplicate node which led to a split, or to some nodes being noise.
def disjoin_if_necessary(edge_list, ps):
    disjointed_hulls_edges = []
    disjointed_pss = []

    # duplicate_edge = None
    # for edge in edge_list:
    #     if edge_list.count(edge) + edge_list.count(toolbox.invert_edge(edge)) > 1:
    #         duplicate_edge = edge
    #         break
    #
    # if duplicate_edge is not None:
    #     occurrences = edge_list.count(edge) + edge_list.count(toolbox.invert_edge(edge))
    #     for i in range(occurrences):
    #         # rotate to edge and then subchain to edge
    #         sub_hull = toolbox.subchain_by_edge(edge_list, duplicate_edge, duplicate_edge, check_symmetrical_edges=True)
    #         sub_hull = sub_hull[:-1]
    #         if len(sub_hull) >= 3:
    #             disjointed_pss.append(toolbox.find_points_in_hull(ps, sub_hull))
    #             disjointed_hulls_edges.append(sub_hull)

    duplicate_node = None
    nodes = toolbox.edge_list_to_hull(edge_list)
    for node in nodes:
        if nodes.count(node) > 1:
            duplicate_node = node
            break

    if duplicate_node is not None:
        for i in range(nodes.count(duplicate_node)):
            sub_hull = toolbox.subchain(edge_list, duplicate_node, duplicate_node)
            if len(sub_hull) > 3:
                disjointed_pss.append(toolbox.find_points_in_hull(ps, sub_hull))
                disjointed_hulls_edges.append(sub_hull)

    was_disjoined = duplicate_node is not None
    if len(disjointed_hulls_edges) > 0:
        return disjointed_hulls_edges, disjointed_pss, was_disjoined
    else:
        return [edge_list], [ps], was_disjoined


def shrink_hull(current_hull_edges, left_hull_edges, right_hull_edges, split_index, longest_edge):

    new_hull_edges = []
    new_hull_edges.extend(toolbox.subchain(left_hull_edges, longest_edge[0], split_index))
    new_hull_edges.extend(toolbox.subchain(right_hull_edges, split_index, longest_edge[1]))
    new_hull_edges.extend(toolbox.subchain_by_edge(current_hull_edges, longest_edge, longest_edge,
                                                   check_symmetrical_edges=False)[:-1])
    #new_hull_edges.extend(toolbox.subchain(current_hull_edges, longest_edge[1], longest_edge[0]))

    return new_hull_edges



def simplify(ps, edges, simplification=0.5, verbose=False):
    desired_number_of_edges = len(edges) * (1 - simplification)
    if verbose:
        print('simplifying ' + str(len(edges) - math.ceil(desired_number_of_edges)) + ' edges...')

    current_edges = edges[:]
    while len(current_edges) > desired_number_of_edges and len(current_edges) > 2:
        longest_edge, longest_edge_length = toolbox.get_longest_edge(ps, current_edges, rank=len(current_edges)-1)
        right_edge_idx = (current_edges.index(longest_edge) + 1) % len(current_edges)
        left_edge_idx = (current_edges.index(longest_edge) - 1) % len(current_edges)
        right_edge = current_edges[right_edge_idx]
        left_edge = current_edges[left_edge_idx]

        if euclidean(ps[right_edge[0]], ps[right_edge[1]]) < euclidean(ps[left_edge[0]], ps[left_edge[1]]):
            current_edges[right_edge_idx] = (longest_edge[0], right_edge[1])
        else:
            current_edges[left_edge_idx] = (left_edge[0], longest_edge[1])

        current_edges.remove(longest_edge)

    return current_edges
