__author__ = 'marco'

import numpy as np
from scipy.spatial import Voronoi
from shapely.geometry import Point, Polygon
from shapely.prepared import prep


import hierarchicalDBSCAN.Clash.concave_hull as concave_hull
import hierarchicalDBSCAN.Clash.toolbox as toolbox


###################
# Skeletonization #
###################

# medial_axis()
#
# Computes a medial axis of a given dataset by shrinking a convex hull until it fits the dataset, and putting the
# medial axis through the resulting concave hull. If necessary, the algorithm will split the dataset into several
# subsets, resulting in an adaptive clustering.
#
# Parameters:
# ps                    ... set of points, as numpy array.
# termination_criterion ... used to determine at which edge length to stop the algorithm.
#                           if 'delauney', the edge length statistics of the delauney triangulations are used.
#                           if 'nearest_neighbor', the distance statistics of the n.n. of each point are used.
#                           if not explicitly set, 'delauney' will be used.
# percentile            ... the percentile used on the delauney (or nearest neighbor) statistics.
# perc_multiplier       ... is multiplied with the <percentile>th percentile of the delauney (or nn) statistics
# adaptive_termination  ... if True, the edge length threshold at which the algorithm stops will be computed for
#                           each found subset (cluster) separately. If False, the threshold of the original dataset
#                           will be used on each found subset, too.
# border_samples        ... ##
#
# Returns:
#  - borders   ... list of edge chains, like [[(idx1, idx2), ...], ...]
#  - skeletons ... list of edge chains, like ##
#  - polygons  ... list of polygons
def skeletonize(ps,
                termination_criterion='delauney',
                percentile=0.75,
                perc_multiplier=2,
                adaptive_termination=True,
                stop_at_disjoin=True,
                simplification=0,
                verbose=False):

    borders, _ = concave_hull.get_concave_hull(ps=ps,
                                                     termination_criterion=termination_criterion,
                                                     percentile=percentile,
                                                     perc_multiplier=perc_multiplier,
                                                     adaptive_termination=adaptive_termination,
                                                     stop_at_disjoin=stop_at_disjoin,
                                                     plot_steps=False,
                                                     plot_steps_after=0,
                                                     verbose=verbose)

#    borders = filter(lambda edges: len(edges) > 2, borders)

    polygons = []
    skeletons = []
    simplified_border_edges = []
    for border_edges in borders:

        border_edges = concave_hull.simplify(ps, border_edges, simplification, verbose)
        if len(border_edges) >= 3:
            simplified_border_edges.append(border_edges)

            border_polygon = get_polygon(ps, border_edges)
            polygons.append(border_polygon)

            #skeleton = get_medial_axis(ps, border_edges, border_polygon)
            #skeletons.append(np.asarray(skeleton))

    return simplified_border_edges, skeletons, polygons


###########
# Voronoi #
###########

def get_polygon(ps, border_edges):
    poly = []
    for edge in border_edges:
        poly.append((ps[edge[0], 0], ps[edge[0], 1]))
    border_polygon = Polygon(poly)
    # border_polygon = border_polygon.simplify(0.075)
    return border_polygon



def sample_on_edges(ps, edges, n_samples):
    samples = []
    lambdas = np.linspace(0.0, 1.0, num=n_samples)
    for edge in edges:
        samples += [l*ps[edge[0]] + (1-l)*ps[edge[1]] for l in lambdas]
    return np.asarray(samples)
