from sklearn.metrics.cluster.supervised import contingency_matrix
import numpy as np

__author__ = 'Alex'

def entropy(labels_true, labels_pred):
    """ return entropy score of true and clustered labeling

    :param labels_true: true labeling of a dataset
    :param labels_pred: predicted labeling of a dataset
    :return: entropy score of the dataset
    """
    n = len(labels_true)
    contingency = contingency_matrix(labels_true, labels_pred)

    entropy_sum = 0
    for i in range(contingency.shape[0]):
        n_i = sum(contingency[0, :])
        for j in range(contingency.shape[1]):
            n_ij_div_n_i = contingency[i,j] / n_i
            if n_ij_div_n_i:
                entropy_sum += (n_i/n)*(n_ij_div_n_i*np.log2(n_ij_div_n_i))
    return np.fabs(-entropy_sum)    # np.fabs is necessary for catching -0.0


def purity(labels_true, labels_pred):
    """ return purity score of true and clustered labeling

    :param labels_true: true labeling of a dataset
    :param labels_pred: predicted labeling of a dataset
    :return: purity score of the dataset
    """
    n = len(labels_true)
    contingency = contingency_matrix(labels_true, labels_pred)

    purity_sum = 0
    for i in range(contingency.shape[0]):
        n_i = sum(contingency[i, :])
        max_n_ij = max(contingency[i,:])
        purity_sum += (n_i/n)*(max_n_ij/n_i)
    return purity_sum


def fmeasure(labels_true, labels_pred):
    """ return F-measure score of true and clustered labeling

    :param labels_true: true labeling of a dataset
    :param labels_pred: predicted labeling of a dataset
    :return: F-measure score of the dataset
    """
    n = len(labels_true)
    contingency = contingency_matrix(labels_true, labels_pred)

    fmeasure_sum = 0
    for j in range(contingency.shape[1]):
        n_j = sum(contingency[:, j])
        max_i = 0
        for i in range(contingency.shape[0]):
            n_i = sum(contingency[i,:])
            n_ij = contingency[i,j]

            value = ((n_ij*n_ij)/(n_i*n_j))/((n_ij/n_i)+(n_ij/n_j))
            if value > max_i:
                max_i = value

        fmeasure_sum += (n_j/n)*(2*max_i)
    return fmeasure_sum

if __name__ == "__main__":
    x = [1, 1, 2, 2, 3, 3]
    y = [1, 2, 3, 4, 5, 6]

    print("Entropy(x,y) = ", entropy(x, y))
    print("Entropy(y,x) = ", entropy(y, x))

    print("Purity(x,y) = ", purity(x, y))
    print("Purity(y,x) = ", purity(y, x))

    print("F-Measure(x,y) = ", fmeasure(x, y))
    print("F-Measure(y,x) = ", fmeasure(y, x))
