__author__ = 'Alex'

from hierarchicalDBSCAN.ClusteringAlgorithms.hierarchicalDBScan import *


class epsHierarchicalDBScan(hierarchicalDBScan):
    """ Clustergraph based implementation of eps
    """

    def __init__(self, eps, minPts = 1):
        self.configure(eps, minPts)


    def configure(self, eps, minPts = 1):
        """ configure clustering object for the next run of fit()

        :param eps: fixed eps value
        :param minPts: minimal level of the hierarchy to be generated
        :return:
        """
        self.eps = eps
        self.minPts = minPts
        self.clustered_dataset = None
        self.clustergraph = None
        self.last_prediction_type = None
        self.simplified_clustergraph = None


    def get_clusterings(self):
        """
        :return: generator object for clustering per hierarchy level
        """
        return labels_for_cut_height_generator(len(self.clustered_dataset), self.simplified_clustergraph, report_heights=True)


    def fit(self, dataset, dist_mat = None):
        """ create clustergraph of dataset

        :param dataset: np array of points
        """
        self.simplified_clustergraph = None
        self.clustered_dataset = dataset
        UNIQUE_NODE_NAME = 0

        if dist_mat is None:
            distance_matrix = pairwise_distances(dataset)
        else:
            distance_matrix = dist_mat
        N = len(dataset)

        linked_Points = {i: np.argwhere(distance_matrix[i,:]<=self.eps).flatten().tolist() for i in range(N)}
        #linked_Points = {i: [item for sublist in linked_Points[i] for item in sublist] for i in range(N)}
        core_minPts = {i:len(linked_Points[i])-1 for i in linked_Points}
        logging.debug("linked Pts", linked_Points)
        logging.debug("core Pts:{0}".format(core_minPts))

        g_list=[list(g) for _, g in groupby(sorted(core_minPts.items(), key=itemgetter(1), reverse=True), lambda x :x[1])]
        dendrogram_levels = [(k[0],list(i)) for i,k in [zip(*i) for i in g_list]]
        logging.debug("g_list: {0}".format(g_list))
        logging.debug("dendrogram_levels: {0}".format(dendrogram_levels))

        added_core_points = set()
        self.clustergraph = nx.DiGraph()
        current_clusters = []


        while dendrogram_levels:
            logging.debug("current clusters: {0}".format(current_clusters))
            current_level = dendrogram_levels.pop(0)

            #add new cores
            new_cores = current_level[1]
            current_level = current_level[0]

            #create one all-including node
            if current_level == 0:
                new_cluster = {"nodes": set(range(N)), "cores": set(range(N)), "name": UNIQUE_NODE_NAME}
                UNIQUE_NODE_NAME += 1
                logging.debug("create final node")
                for c in range(len(current_clusters))[::-1]:
                    self.clustergraph.add_edge(new_cluster["name"], current_clusters[c]["name"])
                    self.clustergraph.node[new_cluster["name"]]["height"] = N - current_level
                    self.clustergraph.node[new_cluster["name"]]["cores"]  = new_cluster["cores"]
                    self.clustergraph.node[new_cluster["name"]]["nodes"]  = new_cluster["nodes"]
                    self.clustergraph.node[new_cluster["name"]]["name"]  = new_cluster["name"]
                    logging.debug("add edge between {0} and {1}, cause of {2}".format(new_cluster["name"], current_clusters[c]["name"], "final node"))

                    #delete all old clusters
                    current_clusters.pop(c)

            else:
                added_core_points.update(new_cores)
                #logging.debug("cores at level {0}: {1}".format(current_level, added_core_points))


                #new core links
                for new_core in new_cores:

                    #get all the clusters the core is in
                    new_cluster = {"nodes": set([new_core]), "cores":set([new_core]), "name":UNIQUE_NODE_NAME}
                    UNIQUE_NODE_NAME += 1

                    cluster_ids = [c for c in range(len(current_clusters)) if new_core in current_clusters[c]["nodes"]]
                    cluster_core_ids = [c for c in range(len(current_clusters)) if new_core in current_clusters[c]["cores"]]
                    if cluster_ids:
                        #merge all clusters the node is in
                        logging.debug("merge clusters {0}, cause of {1}".format(cluster_ids, new_core))
                        for c in cluster_ids:
                            new_cluster["nodes"].update(current_clusters[c]["nodes"])
                            new_cluster["cores"].update(current_clusters[c]["cores"])


                    #additionally add all nodes linked from the new core
                    while len(linked_Points[new_core]) > 0:
                        #process the next link
                        #print("linked_Points", linked_Points[new_core])
                        target = linked_Points[new_core].pop(0)
                        if target == new_core:
                            logging.debug("continue")
                            continue

                        #print("target", target)
                        if not target in new_cluster["nodes"]:
                            new_cluster["nodes"].add(target)
                            if core_minPts[target] >= current_level:
                                #add triple[1] straight to corelist if its minPts is reached
                                new_cluster["cores"].add(target)
                            logging.debug("node {0} added to new cluster {1}, by {2}". format(target, new_cluster["nodes"], new_core))


                    #node was previously added as core in the same time step
                    #remove edges from clusters where it is already a core and replace the source node with the current new cluster
                    if cluster_core_ids:
                        logging.debug("core_ids {0}".format(cluster_core_ids))
                        logging.debug("c_ids {0}".format(cluster_ids))
                        for c in cluster_core_ids:
                            for v in self.clustergraph.neighbors(current_clusters[c]["name"]):
                                self.clustergraph.add_edge(new_cluster["name"], v)
                            self.clustergraph.add_node(new_cluster["name"])
                            self.clustergraph.node[new_cluster["name"]]["height"] = N-current_level
                            self.clustergraph.node[new_cluster["name"]]["cores"]  = new_cluster["cores"]
                            self.clustergraph.node[new_cluster["name"]]["nodes"]  = new_cluster["nodes"]
                            self.clustergraph.node[new_cluster["name"]]["name"]  = new_cluster["name"]
                            self.clustergraph.remove_node(current_clusters[c]["name"])
                            #current_clusters.pop(c)

                        for c in sorted([cluster for cluster in cluster_ids if cluster not in cluster_core_ids], reverse=True):
                            self.clustergraph.add_edge(new_cluster["name"], current_clusters[c]["name"])
                            self.clustergraph.node[new_cluster["name"]]["height"] = N-current_level
                            self.clustergraph.node[new_cluster["name"]]["cores"]  = new_cluster["cores"]
                            self.clustergraph.node[new_cluster["name"]]["nodes"]  = new_cluster["nodes"]
                            self.clustergraph.node[new_cluster["name"]]["name"]  = new_cluster["name"]
                            logging.debug("add edge between {0} and {1}, cause of {2}".format(new_cluster["name"], current_clusters[c]["name"], new_core))
                            #delete all old clusters
                            #current_clusters.pop(c)
                        for c in sorted(cluster_ids, reverse=True):
                            current_clusters.pop(c)


                    else:
                        #add new node to cluster tree
                        if cluster_ids:
                            #add merging edges to the tree
                            for c in sorted(cluster_ids, reverse=True):
                                self.clustergraph.add_edge(new_cluster["name"], current_clusters[c]["name"])
                                self.clustergraph.node[new_cluster["name"]]["height"] = N-current_level
                                self.clustergraph.node[new_cluster["name"]]["cores"]  = new_cluster["cores"]
                                self.clustergraph.node[new_cluster["name"]]["nodes"]  = new_cluster["nodes"]
                                self.clustergraph.node[new_cluster["name"]]["name"]  = new_cluster["name"]
                                logging.debug("add edge between {0} and {1}, cause of {2}".format(new_cluster["name"], current_clusters[c]["name"], new_core))
                                #delete all old clusters
                                current_clusters.pop(c)
                        else:
                            #add new core to the tree
                            self.clustergraph.add_node(new_cluster["name"], {"height":N-current_level,
                                                                        "cores" : new_cluster["cores"],
                                                                        "nodes" : new_cluster["nodes"],
                                                                        "name" : new_cluster["name"]})
                            logging.debug("add node {0}".format(new_cluster["nodes"]))

                    #add new cluster
                    current_clusters.append(new_cluster)


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARNING)

    np.random.seed(17041991)
    #dataset, labels = generate_dataset("blobs", {"n_samples":100,  "random_state":None})
    #dataset, labels = aggregation()
    dataset, labels = standard()
    distance_matrix = pairwise_distances(dataset)

    db = epsHierarchicalDBScan(0.2)
    db.fit(dataset, distance_matrix)
    #db.plot(False)
    #db.plot(True)
    #db.plot_hierarchy_cut()
    db.simplify_clustergraph()

    db.plot_horizontal(init_cut_height=3)
    db.plot_hierarchy_cut()

