__author__ = 'Alex'

import numpy as np
import matplotlib.pyplot as plt
import prettyplotlib as ppl

import sklearn.cluster as cl
from sklearn.neighbors import kneighbors_graph
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import homogeneity_completeness_v_measure
from hierarchicalDBSCAN.Robustness_tests import generate_dataset
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
from hierarchicalDBSCAN.ClusterValidation.ExternalValidation import *

#algorithms
from hierarchicalDBSCAN.ClusteringAlgorithms.DifDenseHDBSCAN import DifDenseDBSCAN
from hierarchicalDBSCAN.ClusteringAlgorithms.alternatingOptimization import aoDBSCAN
from hierarchicalDBSCAN.ClusteringAlgorithms.DifDenseAlphaHDBSCAN import DifDenseAlphaDBSCAN
from hierarchicalDBSCAN.ClusteringAlgorithms.DifDenseClashDescriptorHDBSCAN import DifDenseClashDescriptorDBSCAN
from hierarchicalDBSCAN.Evaluation.Results.csv_to_latex import convert_csv_to_tex

np.random.seed(0)

blobs2d = generate_dataset("blobs", {"n_samples":100, "cluster_std":0.3, "n_features":2, "random_state":None})
blobs1000d = generate_dataset("blobs", {"n_samples":100, "n_features":1000, "random_state":None})
moons = generate_dataset("moons", {"n_samples":300, "noise": 0.075, "random_state":None})
spiral = spiral()
compound = compound()
aggregation = aggregation()
t48k = t48k(size=2000)
flame = flame()
standard = standard()
r15 = r15()
jain = jain()


colors = np.array([x for x in 'bgrcmykbgrcmykbgrcmykbgrcmyk'])
colors = np.hstack([colors] * 20)


clustering_names = ['aoDBSCAN', 'Edge Quantile', 'Alpha Shape']#, 'Clash']# ,'Christians Shape']

base_colors = [(255,187,120),(255,127,14),(174,199,232),(44,160,44),(31,119,180),(255,152,150),(214,39,40),
               (197,176,213),(152,223,138),(148,103,189),(247,182,210),(227,119,194),(196,156,148),(140,86,75),
               (127,127,127),(219,219,141),(199,199,199),(188,189,34),(158,218,229),(23,190,207)]*100


plot_num = 1


datasets = [moons, blobs1000d, r15, compound, flame, jain]
nr_clusters = [2, 3, 15, 6, 1, 2]
data_names = ["Moons", "Blobs",  "R15",  "Compound", "Flame", "Moons-DifDense"]
#datasets = [compound]
#nr_clusters = [6]
#data_names = ["Compound"]




def write_csv(file_path, data, header=None):
    with open(file_path,"w") as f:
        if header is not None:
            f.write(",".join(header) + "\n")
        for line in data: f.write(",".join(line) + "\n")

def draw_labeling(data, labeling, true_labeling=False):
    labeling = np.asarray(labeling)
    if len(set(labeling)) > 1:
        colors = {label: [t/255 for t in base_colors[(i*3)]] for i, label in enumerate(set(labeling))}
        for label in set(labeling):
            filtered_dataset = data[np.where(labeling==label)[0].tolist(), ]
            if label == -1 and not true_labeling:
                ppl.scatter(plt.gca(), filtered_dataset[:, 0], filtered_dataset[:, 1], label=str(label),
                            color = [0.9, 0.9, 0.9], linewidth=1, marker='x', edgecolor='black')
            else:
                #filtered_dataset = dataset[np.where(labeling == label)[0].tolist(),:]
                ppl.scatter(plt.gca(), filtered_dataset[:, 0], filtered_dataset[:, 1], label=str(label),
                            color=colors[label])
    else:
        ppl.scatter(plt.gca(), data[:, 0], data[:, 1], label=str(1),
                    linewidth=1, marker='x', edgecolor='black',
                            color=[0.9, 0.9, 0.9])


for i_dataset, dataset in enumerate(datasets):
    print("dataset", data_names[i_dataset])

    X, y = dataset


    hdbscan_edge_cut = DifDenseDBSCAN(minPts=4)
    hdbscan_alpha_cut = DifDenseAlphaDBSCAN(minPts=5)
    #hdbscan_shape_cut = DifDenseClashDescriptorDBSCAN(minPts=5)

    ao_modsil = aoDBSCAN(opt="modsil")


    clustering_algorithms = [ao_modsil, hdbscan_edge_cut, hdbscan_alpha_cut]


    algo = 0
    results = []
    clustering = []
    for name, algorithm, subplot in zip(clustering_names, clustering_algorithms, range(2,len(clustering_names)+2)):
        print("algorithm", name, "subplot", subplot)
        # predict cluster memberships
        algorithm.fit(X)

        if name == 'Alpha Shape':
            if data_names[i_dataset] == 'Blobs':
                y_pred = [0]*len(y)
            else:
                y_pred = algorithm.predict(X, nr_clusters[i_dataset])
        else:
            if hasattr(algorithm, 'labels_'):
                y_pred = list(algorithm.labels_.astype(np.int))
            else:
                y_pred = algorithm.predict(X)

        ent = entropy(y, y_pred)
        pur = purity(y, y_pred)
        fmeas = fmeasure(y, y_pred)
        vmeas = homogeneity_completeness_v_measure(y, y_pred)[2]
        eval = homogeneity_completeness_v_measure(y, y_pred)
        results.extend((name,ent,pur,fmeas,vmeas))

        clustering.append((name, y_pred))

        # plot
        #plt.subplot(1, len(clustering_names)+1, subplot)
        #draw_labeling(X, y_pred)
        #plt.title(name, size=12)


    fig = plt.figure(figsize=((len(clustering_names)+1)*4, 4))
    plt.subplot(1, len(clustering_names)+1, 1)
    draw_labeling(X, y, True)
    plt.title("True Clustering", size=20)
    for i, c in enumerate(clustering):
        plt.subplot(1, len(clustering_names)+1, i + 2)
        draw_labeling(X, c[1])
        plt.title(c[0], size=20)
    fig.tight_layout()
    fig.subplots_adjust(hspace=0.3)
    fig.savefig('../../../ssci2016/images/'+ data_names[i_dataset] +'_new.png')
    plt.close(fig)


    #test = np.rec.fromarrays((np.array(data_names), evaluation), names=("algorithm", "data"))
    dt=np.dtype('|S10,f4,f4,f4')
    evaluation = np.array(results)
    evaluation = evaluation.reshape(len(clustering_algorithms),5)
    print(evaluation)
    write_csv("../../../ssci2016/results/"+data_names[i_dataset]+"_new_eval.csv", evaluation, header=["algorithm","entropy", "purity", "fmeasure", "vmeasure"])
    #convert_csv_to_tex(data_names[i_dataset])
