__author__ = 'Alex'

"""
compute the mean and stddev of 100 data sets and plot mean vs stddev.
When you click on one of the mu, sigma points, plot the raw data from
the dataset that generated the mean and stddev
"""
import numpy as np
import matplotlib.pyplot as plt
import matplotlib
from sklearn.cluster import DBSCAN
from sklearn.metrics.pairwise import pairwise_distances
from hierarchicalDBSCAN.ClusterValidation.InternalValidation import get_modified_silhouette_score
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
import prettyplotlib as ppl
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.patches import Rectangle


base_colors = [(255,187,120),(255,127,14),(174,199,232),(44,160,44),(31,119,180),(255,152,150),(214,39,40),
               (197,176,213),(152,223,138),(148,103,189),(247,182,210),(227,119,194),(196,156,148),(140,86,75),
               (127,127,127),(219,219,141),(199,199,199),(188,189,34),(158,218,229),(23,190,207)]*100


def forceAspect(ax,aspect=1):
    im = ax.get_images()
    extent =  im[0].get_extent()
    ax.set_aspect(abs((extent[1]-extent[0])/(extent[3]-extent[2]))/aspect)


def draw_labeling(data, labeling):
    labeling = np.asarray(labeling)
    if len(set(labeling)) > 1:
        colors = {label: [t/255 for t in base_colors[(i*3)]] for i, label in enumerate(set(labeling))}
        for label in set(labeling):
            filtered_dataset = data[np.where(labeling==label)[0].tolist(), ]

            #filtered_dataset = dataset[np.where(labeling == label)[0].tolist(),:]
            ppl.scatter(plt.gca(), filtered_dataset[:, 0], filtered_dataset[:, 1], label=str(label),
                        color=colors[label], linewidth=1, edgecolor='black')
    else:
        ppl.scatter(plt.gca(), data[:, 0], data[:, 1],
                    color='black', linewidth=1, edgecolor='black', alpha=1)


def draw_dbscan_heatmap(dataset, min_eps, max_eps, min_minPts, max_minPts, steps_eps=10, steps_minPts=10):
    X, y = dataset
    dist_mat = pairwise_distances(X)

    eps_stops = np.linspace(min_eps, max_eps, num=steps_eps+1)
    shift_eps_stop = (eps_stops[1] - eps_stops[0])/2
    minPts_stops = np.linspace(min_minPts, max_minPts, num=min(steps_minPts+1, max_minPts-min_minPts))
    shift_minPts_stop = (minPts_stops[1] - minPts_stops[0])/2

    heatmap = []
    for eps in eps_stops[:-1]:
        current_list = []
        for minPts in minPts_stops[:-1]:
            try:
                db = DBSCAN(eps=eps+shift_eps_stop, min_samples=minPts+shift_minPts_stop, metric="precomputed").fit(dist_mat)
                score = get_modified_silhouette_score(dist_mat, db.labels_, minPts)
            except ValueError:
                score = 0

            #print("eps={0:.2f}, minPts={1:.2f}, score{2}".format(eps, minPts, score))
            current_list.append(score)
        current_list.reverse()
        heatmap.append(current_list)
    #print(heatmap)
    heatmap = np.array(heatmap)
    heatmap = heatmap.transpose()
    #print(heatmap)

    fig = plt.figure(figsize=(15, 6))
    plt.subplot(1, 2, 1)
    ax = plt.gca()
    extent = [eps_stops[0], eps_stops[-1], minPts_stops[0], minPts_stops[-1]]
    im = plt.imshow(heatmap, interpolation="nearest", extent=extent, picker=True)
    forceAspect(ax,aspect=1)
    ax.set_title('click on a cell to show clustering')
    rect = Rectangle((0,0), 0, 0, fill=False, lw=3)
    ax.add_patch(rect)

    plt.xlabel('eps',{'fontsize':16})
    plt.ylabel('minPts',{'fontsize':16})

    #divider = make_axes_locatable(ax)
    #cax = divider.append_axes("right", size="5%", pad=0.05)
    #plt.colorbar(im, cax=cax)
    #plt.colorbar()
    plt.colorbar(im,fraction=0.046, pad=0.04)



    def on_pick(event):
        eps = event.mouseevent.xdata
        minPts = event.mouseevent.ydata

        rounded_eps = eps_stops[0]
        for stop in eps_stops:
            if stop > eps:
                break
            else:
                rounded_eps = stop

        rounded_minPts = minPts_stops[0]
        for stop in minPts_stops:
            if stop > minPts:
                break
            else:
                rounded_minPts = stop

        plt.subplot(1, 2, 1)
        rect.set_width(eps_stops[1]-eps_stops[0])
        rect.set_height(minPts_stops[1]-minPts_stops[0])
        rect.set_xy((rounded_eps, rounded_minPts))
        plt.gca().figure.canvas.draw()

        plt.subplot(1, 2, 2)
        plt.cla()
        try:
            db = DBSCAN(eps=rounded_eps+shift_eps_stop, min_samples=rounded_minPts+shift_minPts_stop, metric="precomputed").fit(dist_mat)
            labels = db.labels_
            #score = get_modified_silhouette_score(dist_mat, db.labels_, rounded_minPts)
            #print("eps={0:.2f}, minPts={1:.2f}, score{2}".format(rounded_eps, rounded_minPts, score))
        except ValueError:
            labels = [-1]*len(X)

        draw_labeling(X, labels)
        ax = plt.gca()
        ax.set_title("DBSCAN (eps={0:.2f}, minPts={1:.2f})".format(rounded_eps+shift_eps_stop, rounded_minPts+shift_minPts_stop))
        plt.draw()

        return True


    fig.canvas.mpl_connect('pick_event', on_pick)

    plt.subplot(1, 2, 2)
    ppl.scatter(X[:,0], X[:,1])

    plt.show()

if __name__ == "__main__":
    dataset = standard()
    draw_dbscan_heatmap(dataset, 0.025, 0.475, 1, 20, steps_eps=10, steps_minPts=10)